
//			DEGAS command:


#ifndef DEGAS_COMMAND_VER
#define DEGAS_COMMAND_VER					1.0.1.1


#define DEGAS_FLAG_HEADER					0x0001
#define DEGAS_FLAG_FIX_QUOTES				0x0002
#define DEGAS_FLAG_TRIM_LINES				0x0008
#define DEGAS_FLAG_NO_LEADING_BLANKS		0x0010
#define DEGAS_FLAG_NO_TRAILING_BLANKS		0x0020
#define DEGAS_FLAG_REPLACE_ALL_WHITESPACE	0x0040
#define DEGAS_FLAG_LINE_NUMBERS				0x0100
#define DEGAS_FLAG_IN_QUOTES				0x0200





const wchar_t DegasHelpText[] = L"Remove excess spaces and blank lines from text.\n\
\n\
DEGAS /B:n /CP:n /E:n /H /L /P /R /S /T /W filename....\r\n\
\n\
\t/B:   maximum whitespace characters\n\
\t/CP:  interpret OEM text as code page n\n\
\t/E:   maximum blank lines\n\
\t/H    display filenames\n\
\t/L    line numbers\n\
\t/P    page output\n\
\t/R    remove all blank lines from start and end of file\n\
\t/S    search in subdirectories for matching files\n\
\t/T    trim leading and trailing whitespace\n\
\t/W    convert all whitespace characters to ASCII spaces\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into DEGAS.\n";



struct _DegasArgs {
	unsigned int MaxBlankLines;
	unsigned int MaxSpacesAfterTerminal;
	unsigned int MaxSpacesAfterOther;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};




BOOL Degas_TerminalPunct( wchar_t ch )
{
	return ( ( ch == '.' ) || ( ch == '!' ) || ( ch == '?' ) );
}



BOOL Degas_EndQuoteChar( wchar_t ch )
{
	return ( ( ch == '"' ) || ( ch == '\'' ) || ( ch == 0x2019 ) || ( ch == 0x201d ) );
}



int DegasLine( LPTSTR Line, unsigned int MaxSpacesAfterTerminal, unsigned int MaxSpacesAfterOther )
{
	if ( !Line )
		return -1;


	wchar_t temp[BIG_LINE_LEN]		= L"";
	wchar_t *ch						= Line;
	unsigned int SpacesCount		= 0;
	unsigned int j					= 0;
	BOOL Terminal					= FALSE;
	int rv							= 0;


	for ( ch = Line; *ch; ch++ ) {

		if ( iswspace( *ch ) ) {
			SpacesCount++;

			if ( Terminal && ( SpacesCount > MaxSpacesAfterTerminal ) ) {
				rv++;
				continue;
			}

			if ( !Terminal && ( SpacesCount > MaxSpacesAfterOther ) ) {
				rv++;
				continue;
			}
		}
		else {
			SpacesCount = 0;

			if ( Degas_TerminalPunct( *ch ) )
				Terminal = TRUE;
			else if ( Degas_EndQuoteChar( *ch ) )
				;
			else
				Terminal = FALSE;
		}

		if ( j < ( BIG_LINE_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';
		}
	}

	if ( rv )
		wcscpy_s( Line, BIG_LINE_LEN, temp );

	return rv;
}



int Degas_ReplaceAllWhitespace( LPTSTR Line )
{
	if ( !Line )
		return -1;


	int rv			= 0;


	for ( wchar_t *ch = Line; *ch; ch++ ) {

		if ( iswspace( *ch ) && ( *ch != 0x20 ) )
			*ch = 0x20, rv++;
	}

	return rv;
}



BOOL Degas_IsBlankLine( LPTSTR Line )
{
	if ( !Line )
		return FALSE;

	wchar_t *ch			= Line;


	while ( *ch ) {
		if ( !iswspace( *ch ) )
			return FALSE;

		ch++;
	}

	return TRUE;
}



int Degas_RemoveTrailingBlankLines( _MMFile *MMFile )
{
	if ( !MMFile )
		return -1;


	size_t OldIndex		= MMFile->Index;
	size_t Index		= MMFile->FileSize;
	size_t NewIndex		= 0;
	BOOL RemovedAny		= FALSE;
	int ch				= 0;


	while ( Index > MMFile->BOMSize ) {

		ch = MMPrevChar( MMFile, Index, &NewIndex );
		if ( ch < 0 )
			break;

		if ( !_MMIsEOL( ch ) && !iswspace( ch ) )
			break;

		Index = NewIndex;
		RemovedAny = TRUE;
	}

	if ( !RemovedAny )
		return 0;

	if ( Index <= MMFile->BOMSize ) {
		MMFile->FileSize = MMFile->BOMSize;
		return 2;
	}

	MMFile->Index = Index;

	do {
		ch = MMPeekAhead( MMFile, 0 );

		if ( ( ch > 0 ) && iswspace( ch ) && !_MMIsEOL( ch ) )
			MMGetChar( MMFile );
		else
			break;
	}
	while ( TRUE );

	ch = MMPeekAhead( MMFile, 0 );
	if ( _MMIsEOL( ch ) ) {
		MMGetChar( MMFile );
		if ( MMPeekAhead( MMFile, 0 ) == _MM2PartEOL( ch ) )
			MMGetChar( MMFile );
	}

	if ( MMFile->Index < MMFile->FileSize ) {
		MMFile->FileSize = MMFile->Index;
		MMFile->Index = OldIndex;
		return 1;
	}

	MMFile->Index = OldIndex;
	return 0;
}



int DegasFile( LPCTSTR Filename, int Flags, void *MoreArgs )
{
	if ( !Filename || !MoreArgs )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_MMFile MMFile;
	wchar_t LineBuffer[BIG_LINE_LEN]	= L"";
	_DegasArgs *Args					= (_DegasArgs*) MoreArgs;

	int GetLineFlags					= 0;
	BOOL BlankLine						= FALSE;
	BOOL NonblankLineFound				= FALSE;
	unsigned int LineNumber				= 0;
	unsigned int BlankLinesCount		= 0;
	int rv								= 0;


	if ( Args->MaxBlankLines == 0 ) {
		GetLineFlags |= MMGL_FLAG_SKIP_BLANKS;
		Flags &= ~( DEGAS_FLAG_NO_LEADING_BLANKS | DEGAS_FLAG_NO_TRAILING_BLANKS );
	}


	rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, Args->CodePage, Args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	if ( Flags & DEGAS_FLAG_NO_TRAILING_BLANKS )
		Degas_RemoveTrailingBlankLines( &MMFile );


	if ( Flags & DEGAS_FLAG_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, GetLineFlags ) )
			break;

		if ( Flags & DEGAS_FLAG_TRIM_LINES )
			TrimLine( LineBuffer );

		BlankLine = Degas_IsBlankLine( LineBuffer );

		if ( BlankLine ) {
			BlankLinesCount++;
			if ( BlankLinesCount > Args->MaxBlankLines )
				continue;

			if ( ( Flags & DEGAS_FLAG_NO_LEADING_BLANKS ) && !NonblankLineFound )
				continue;
		}
		else {
			BlankLinesCount = 0;
			NonblankLineFound = TRUE;

			if ( Flags & DEGAS_FLAG_REPLACE_ALL_WHITESPACE )
				Degas_ReplaceAllWhitespace( LineBuffer );

			if ( Flags & DEGAS_FLAG_FIX_QUOTES )
				FixQuotes( LineBuffer, BIG_LINE_LEN );
		}

		if ( Flags & DEGAS_FLAG_LINE_NUMBERS ) {
			LineNumber++;
			Printf( L"%4u : ", LineNumber );
		}

		if ( ( Args->MaxSpacesAfterTerminal < UINT_MAX ) || ( Args->MaxSpacesAfterOther < UINT_MAX ) )
			DegasLine( LineBuffer, Args->MaxSpacesAfterTerminal, Args->MaxSpacesAfterOther );


		if ( Flags & DEGAS_FLAG_IN_QUOTES )
			QPuts( L"\"" );

		QPuts( LineBuffer );

		if ( Flags & DEGAS_FLAG_IN_QUOTES )
			QPuts( L"\"" );

		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}


	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort :

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}




DLLExports int WINAPI degas( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= 0;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	BOOL AnyFixSpecified	= FALSE;
	_DegasArgs Args			= { UINT_MAX, UINT_MAX, UINT_MAX, 0, 0 };


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( DegasHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( DEGAS_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/B", 2 ) ) {
				wchar_t *sep1	= NULL;
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &(Args.MaxSpacesAfterTerminal), &sep1 ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad MaxSpaces value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( *sep1 == '\0' )
					Args.MaxSpacesAfterOther = Args.MaxSpacesAfterTerminal;
				else if ( ( *sep1 == ',' ) || ( *sep1 == ';' ) ) {
					sep1++;

					if ( ParseInt( sep1, &(Args.MaxSpacesAfterOther), NULL ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad MaxSpaces value: \"%s\"\r\n", arg[i] );
						goto ErrorExit;
					}
				}
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error (expected a comma): \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				AnyFixSpecified = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/E" ) )
				Args.MaxBlankLines = 0, AnyFixSpecified = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/E", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &(Args.MaxBlankLines), NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad MaxBlankLines value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				AnyFixSpecified = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= DEGAS_FLAG_HEADER;
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				Flags |= DEGAS_FLAG_LINE_NUMBERS;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= DEGAS_FLAG_FIX_QUOTES;
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Flags |= ( DEGAS_FLAG_NO_LEADING_BLANKS | DEGAS_FLAG_NO_TRAILING_BLANKS ), AnyFixSpecified = TRUE;
			else if ( !_wcsicmp( arg[i], L"/RS" ) )
				Flags |= DEGAS_FLAG_NO_LEADING_BLANKS, AnyFixSpecified = TRUE;
			else if ( !_wcsicmp( arg[i], L"/RE" ) )
				Flags |= DEGAS_FLAG_NO_TRAILING_BLANKS, AnyFixSpecified = TRUE;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Flags |= DEGAS_FLAG_TRIM_LINES, AnyFixSpecified = TRUE;
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				Flags |= DEGAS_FLAG_REPLACE_ALL_WHITESPACE, AnyFixSpecified = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				Flags |= DEGAS_FLAG_IN_QUOTES;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( !AnyFixSpecified ) {
		Args.MaxBlankLines			= 1;
		Args.MaxSpacesAfterTerminal	= 2;
		Args.MaxSpacesAfterOther	= 1;
	};



	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );

	if ( Flags & DEGAS_FLAG_FIX_QUOTES )
		InitializeFixQuotesVars();

	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );



	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = DegasFile( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, DegasFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit :

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort :

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	DEGAS_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				DegasFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
