
//			FFIELDS command:


#ifndef FFIELDS_COMMAND_VER
#define FFIELDS_COMMAND_VER			1.0.1.1


#define FF_HEADER					0x0001
#define FF_STRIP_EQUALS				0x0040
#define FF_DEBUG					0x0080
#define FF_EXPAND					0x0100
#define FF_EXPAND_SAFELY			0x0200
#define FF_DETAB					0x0400
#define FF_FMT_LEFT					0
#define FF_FMT_RIGHT				1
#define FF_FMT_CENTER				2
#define FF_FMT_TRUNCATE				3



const wchar_t FFieldsHelpText[] = L"Read a file and print fields in a specified format.\n\
\n\
FFIELDS /C /CP:n /E /F:\"format\" /H /K:n /L /P /S /T /W /X filename....\r\n\
\n\
\t/C    separate fields at commas\n\
\t/CP:  interpret OEM text as code page n\n\
\t/E    separate fields at first unquoted equals sign\n\
\t/F:   format string; see below\n\
\t/H    display filenames\n\
\t/K:   output line width\n\
\t/L    insert line numbers\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\t/T    separate fields at tabs\n\
\t/W    separate fields at whitespace\n\
\t/X    perform variable expansion on each line\n\
\n\
The format string may contain $n to print field n, or $n=wf to print\n\
field n truncated to length w; the final letter is L to left-justify, \n\
R to right-justify, C to center, or T to truncate the field.  $4=10L\n\
would print field 4, left-justified in 10 columns.  Fields are\n\
numbered starting from 0.\n\
\n\
Input filenames may be specified on the command line, or text may\n\
be piped or redirected into FFIELDS.\n";



struct FileFieldsArgs {
	LPTSTR Format;
	unsigned int ParseFlags;
	LPTSTR LineString;
	int Columns;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



int FFAppendArg( LPTSTR String, int ArgNum, int ArgLen, int ArgFmt, int LineWidth )
{
	if ( !String )
		return -1;

	if ( ( ArgNum >= MAX_ARGS ) || ( ArgNum >= numargs ) )
		return 0;

	if ( ArgLen == 0 ) {
		if ( arg[ArgNum] )
			wcsncat_s( String, MAX_ARG_LEN, arg[ArgNum], _TRUNCATE );

		return 0;
	}

	if ( ArgLen == UINT_MAX ) {
		wchar_t *ch = String;

		while ( wcschr( ch, CR ) != NULL )
			ch = wcschr( ch, CR ) + 1;

		while ( wcschr( ch, LF ) != NULL )
			ch = wcschr( ch, LF ) + 1;

		int len = (int) wcslen( ch );

		if ( len < ( LineWidth - 1 ) )
			ArgLen = LineWidth - len - 1;
		else
			ArgLen = 18;
	}

	wchar_t temp[MAX_ARG_LEN] = L"";

	int leftpad = 0, rightpad = 0;

	if ( arglen[ArgNum] < ArgLen ) {
		int diff = ArgLen - arglen[ArgNum];

		if ( ArgFmt == FF_FMT_LEFT )
			rightpad = diff;
		else if ( ArgFmt == FF_FMT_RIGHT )
			leftpad = diff;
		else if ( ArgFmt == FF_FMT_CENTER ) {
			leftpad = diff / 2;
			rightpad = diff - leftpad;
		}
	}


	if ( leftpad ) {
		for ( int i = 0; i < leftpad; i++ )
			temp[i] = SPACE;

		temp[leftpad] = '\0';
	}

	if ( arg[ArgNum] )
		wcsncat_s( temp, MAX_ARG_LEN, arg[ArgNum], _TRUNCATE );

	if ( rightpad ) {
		int l = (int) wcslen( temp );
		for ( int i = 0; i < rightpad; i++ ) {
			if ( ( l + i ) < ( MAX_ARG_LEN - 1 ) )
				temp[l + i] = SPACE;

			if ( ( l + rightpad ) < MAX_ARG_LEN )
				temp[l + rightpad] = '\0';
			else
				temp[MAX_ARG_LEN - 1] = '\0';
		}
	}

	wcsncat_s( String, MAX_ARG_LEN, temp, ArgLen );
	return 0;
}



int FileFields( LPCTSTR Filename, int DisplayFlags, void *FieldsArgs )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t LineBuffer[MAX_ARG_LEN] = L"";
	wchar_t temp[MAX_ARG_LEN] = L"";
	unsigned int LineNumber = 0;
	wchar_t OldLineVar[MAX_ARG_LEN] = L"";
	wchar_t VarName[] = L"_LINE";
	wchar_t ch = 0x00, *sep1 = NULL;
	unsigned int ArgNum = 0, ArgLen = 0, ArgFmt = FF_FMT_LEFT;
	unsigned int AutoLineNumberWidth = 0;

	FileFieldsArgs *MoreArgs = (FileFieldsArgs*) FieldsArgs;
	LPTSTR Format = MoreArgs->Format, LineString = MoreArgs->LineString;
	unsigned int ParseFlags = MoreArgs->ParseFlags;


	_MMFile MMFile;
	int rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;



	CL_HookBreak();
	
	if ( DisplayFlags & FF_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	GetEnvironmentVariable( VarName, OldLineVar, MAX_ARG_LEN );


	if ( Format[0] == '\0' ) {
		int MostArgs = 0;
		int LongestArg[MAX_ARGS];
		int NumLines = 0;

		for ( int i = 0; i < MAX_ARGS; i++ )
			LongestArg[i] = 0;

		while ( MMGetLine( &MMFile, LineBuffer, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) == 0 ) {

			ParseArgs( LineBuffer, ParseFlags );
			if ( numargs > MostArgs )
				MostArgs = numargs;

			if ( ( DisplayFlags & FF_STRIP_EQUALS ) && numargs )
				if ( arg[numargs - 1][0] == '=' )
					arglen[numargs - 1]--;

			for ( int j = 0; j < numargs; j++ )
				if ( arglen[j] > LongestArg[j] )
					LongestArg[j] = arglen[j];

			NumLines++;
		}

		if ( DisplayFlags & FF_DEBUG ) {
			Printf( L"   * Most args = %d\n", MostArgs );

			for ( int i = 0; i < MostArgs; i++ )
				Printf( L"   * Longest arg[%d] = %d\n", i, LongestArg[i] );
		}


		int ColumnsRemaining = MoreArgs->Columns - 1;

		if ( LineString ) {
			if ( LineString[0] ) {
				SetEnvironmentVariable( VarName, L"0" );
				wcscpy_s( temp, MAX_ARG_LEN, LineString );
				ExpandVariables( temp, 1 );
				int l1 = (int) wcslen( temp );

				wsprintf( temp, L"%u", NumLines - 1 );
				SetEnvironmentVariable( VarName, temp );
				wcscpy_s( temp, MAX_ARG_LEN, LineString );
				ExpandVariables( temp, 1 );
				int l2 = (int) wcslen( temp );

				ColumnsRemaining -= ( l2 > l1 ) ? l2 : l1;
			}
			else {
				wsprintf( temp, L"%u", NumLines - 1 );
				AutoLineNumberWidth = (int) wcslen( temp );
				ColumnsRemaining = ColumnsRemaining - AutoLineNumberWidth - 1;
			}
		}

		for ( int ai = 0; ai < MostArgs; ai++ ) {

			if ( ai == ( MostArgs - 1 ) ) {
				if ( LongestArg[ai] < ColumnsRemaining )
					wsprintf( temp, L"$%d=%dL", ai, LongestArg[ai] );
				else
					wsprintf( temp, L"$%d=%dT", ai, ColumnsRemaining );

				wcscat_s( Format, MAX_ARG_LEN, temp );
			}
			else {
				int aw = 18;
				int est = 0;


				if ( aw > ColumnsRemaining )
					aw = ColumnsRemaining;

				est = ( ColumnsRemaining / ( MostArgs - ai ) ) - 2;

				if ( LongestArg[ai] <= est )
					aw = LongestArg[ai];

				if ( aw > LongestArg[ai] )
					aw = LongestArg[ai];

				if ( aw < 3 )
					aw = 3;

				if ( aw > ColumnsRemaining )
					aw = ColumnsRemaining;


				if ( aw > ( ColumnsRemaining - 10 ) ) {
					wsprintf( temp, L"$%d=%dT", ai, ColumnsRemaining );
					wcscat_s( Format, MAX_ARG_LEN, temp );
					break;
				}
				else {
					wsprintf( temp, L"$%d=%dL  ", ai, aw );
					wcscat_s( Format, MAX_ARG_LEN, temp );
					ColumnsRemaining = ColumnsRemaining - aw - 2;
				}
			}
		}

		if ( DisplayFlags & FF_DEBUG )
			Printf( L"   * Format = \"%s\"\n", Format );

		MMFile.Index = MMFile.BOMSize;
	}
	else if ( LineString ) {
		if ( LineString[0] == '\0' ) {
			int NumLines = 0;

			while ( MMGetLine( &MMFile, LineBuffer, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) == 0 )
				NumLines++;

			wsprintf( temp, L"%u", NumLines - 1 );
			AutoLineNumberWidth = (int) wcslen( temp );

			MMFile.Index = MMFile.BOMSize;
		}
	}
	

	while ( MMGetLine( &MMFile, LineBuffer, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) == 0 ) {

		if ( DisplayFlags & FF_DETAB )
			Detab( LineBuffer, 4, SPACE );

		ParseArgs( LineBuffer, ParseFlags );

		if ( ( DisplayFlags & ( FF_EXPAND | FF_EXPAND_SAFELY ) ) == ( FF_EXPAND | FF_EXPAND_SAFELY ) ) {
			SafeChars( LineBuffer );
			for ( int k = 0; k < numargs; k++ )
				SafeChars( arg[k] );
		}

		if ( ( DisplayFlags & FF_STRIP_EQUALS ) && numargs )
			if ( arg[numargs - 1][0] == '=' ) {
				arglen[numargs - 1]--;
				arg[numargs - 1]++;
			}

		wsprintf( temp, L"%u", LineNumber );
		SetEnvironmentVariable( VarName, temp );

		temp[0] = '\0';
		int j = 0;

		if ( LineString ) {
			if ( LineString[0] ) {
				wcscpy_s( temp, MAX_ARG_LEN, LineString );
				ExpandVariables( temp, 1 );
			}
			else if ( AutoLineNumberWidth )
				swprintf( temp, MAX_ARG_LEN - 1, L"%*u ", AutoLineNumberWidth, LineNumber );
			else
				wcscat_s( temp, MAX_ARG_LEN, L" " );

			j = (int) wcslen( temp );
		}


		for ( int i = 0; Format[i]; i++ ) {
			ch = Format[i];

			if ( ch == '$' ) {
				if ( Format[i + 1] == '$' )
					i++;
				else if ( towupper( Format[i + 1] ) == 'N' ) {
					if ( j < ( MAX_ARG_LEN - 1 ) )
						temp[j++] = '\r';

					ch = '\n';
				}
				else if ( towupper( Format[i + 1] ) == 'E' ) {
					ch = ESC;
					i++;
				}
				else if ( towupper( Format[i + 1] ) == 'K' ) {
					ch = '`';
					i++;
				}
				else if ( towupper( Format[i + 1] ) == 'Q' ) {
					ch = QUOTE;
					i++;
				}
				else if ( towupper( Format[i + 1] ) == 'T' ) {
					ch = TAB;
					i++;
				}
				else if ( iswdigit( Format[i + 1] ) ) {
					ArgNum = 0;
					ArgLen = 0;
					ArgFmt = FF_FMT_LEFT;

					if ( ParseInt( Format + i + 1, &ArgNum, &sep1 ) )
						ArgNum = MAX_ARGS;

					if ( *sep1 == '=' ) {

						sep1++;

						if ( *sep1 == '*' ) {
							sep1++;
							ArgLen = UINT_MAX;

							switch ( towupper( *sep1 ) ) {
								case 'L' :
									ArgFmt = FF_FMT_LEFT;
									sep1++;
									break;

								case 'R' :
									ArgFmt = FF_FMT_RIGHT;
									sep1++;
									break;

								case 'C' :
									ArgFmt = FF_FMT_CENTER;
									sep1++;
									break;

								case 'T' :
									ArgFmt = FF_FMT_TRUNCATE;
									sep1++;
									break;
							}
						}
						else if ( ParseInt( sep1, &ArgLen, &sep1 ) == 0 ) {

							if ( ArgLen > ( MAX_ARG_LEN - 1 ) )
								ArgLen = MAX_ARG_LEN - 1;

							switch ( towupper( *sep1 ) ) {
								case 'L' :
									ArgFmt = FF_FMT_LEFT;
									sep1++;
									break;

								case 'R' :
									ArgFmt = FF_FMT_RIGHT;
									sep1++;
									break;

								case 'C' :
									ArgFmt = FF_FMT_CENTER;
									sep1++;
									break;

								case 'T' :
									ArgFmt = FF_FMT_TRUNCATE;
									sep1++;
									break;
							}
						}
					}

					i = (int) ( sep1 - Format - 1 );

					if ( ArgNum < MAX_ARGS ) {
						FFAppendArg( temp, ArgNum, ArgLen, ArgFmt, MoreArgs->Columns );
						j = (int) wcslen( temp );
					}

					ch = '\0';
				}
				else if ( Format[i + 1] == '*' ) {
					ArgLen = 0;
					ArgFmt = FF_FMT_LEFT;

					sep1 = Format + i + 2;

					if ( *sep1 == '=' ) {

						sep1++;

						if ( *sep1 == '*' ) {
							sep1++;
							ArgLen = UINT_MAX;

							switch ( towupper( *sep1 ) ) {
								case 'L' :
									ArgFmt = FF_FMT_LEFT;
									sep1++;
									break;

								case 'R' :
									ArgFmt = FF_FMT_RIGHT;
									sep1++;
									break;

								case 'C' :
									ArgFmt = FF_FMT_CENTER;
									sep1++;
									break;

								case 'T' :
									ArgFmt = FF_FMT_TRUNCATE;
									sep1++;
									break;
							}
						}
						else if ( ParseInt( sep1, &ArgLen, &sep1 ) == 0 ) {

							if ( ArgLen > ( MAX_ARG_LEN - 1 ) )
								ArgLen = MAX_ARG_LEN - 1;

							switch ( towupper( *sep1 ) ) {
								case 'L' :
									ArgFmt = FF_FMT_LEFT;
									sep1++;
									break;

								case 'R' :
									ArgFmt = FF_FMT_RIGHT;
									sep1++;
									break;

								case 'C' :
									ArgFmt = FF_FMT_CENTER;
									sep1++;
									break;

								case 'T' :
									ArgFmt = FF_FMT_TRUNCATE;
									sep1++;
									break;
							}
						}
					}

					i = (int) ( sep1 - Format - 1 );

					if ( ArgLen == 0 )
						wcsncat_s( temp, MAX_ARG_LEN, LineBuffer, _TRUNCATE );
					else {
						if ( ArgLen == UINT_MAX ) {
							wchar_t *ch2 = temp;

							while ( wcschr( ch2, CR ) != NULL )
								ch2 = wcschr( ch2, CR ) + 1;

							while ( wcschr( ch2, LF ) != NULL )
								ch2 = wcschr( ch2, LF ) + 1;

							int len = (int) wcslen( ch2 );

							if ( len < ( MoreArgs->Columns - 1 ) )
								ArgLen = MoreArgs->Columns - len - 1;
							else
								ArgLen = 18;
						}
					}

					wchar_t temp2[MAX_ARG_LEN] = L"";

					int leftpad = 0, rightpad = 0;

					if ( (int) wcslen( LineBuffer ) < ArgLen ) {
						int diff = ArgLen- (int) wcslen( LineBuffer );

						if ( ArgFmt == FF_FMT_LEFT )
							rightpad = diff;
						else if ( ArgFmt == FF_FMT_RIGHT )
							leftpad = diff;
						else if ( ArgFmt == FF_FMT_CENTER ) {
							leftpad = diff / 2;
							rightpad = diff - leftpad;
						}
					}

					if ( leftpad ) {
						for ( int k = 0; k < leftpad; k++ )
							temp2[k] = SPACE;

						temp2[leftpad] = '\0';
					}

					wcsncat_s( temp2, MAX_ARG_LEN, LineBuffer, _TRUNCATE );

					if ( rightpad ) {
						int l = (int) wcslen( temp2 );
						for ( int k = 0; k < rightpad; k++ ) {
							if ( ( l + k ) < ( MAX_ARG_LEN - 1 ) )
								temp2[l + k] = SPACE;

							if ( ( l + rightpad ) < MAX_ARG_LEN )
								temp2[l + rightpad] = '\0';
							else
								temp2[MAX_ARG_LEN - 1] = '\0';
						}
					}

					wcsncat_s( temp, MAX_ARG_LEN, temp2, ArgLen );
					j = (int) wcslen( temp );
					ch = '\0';
				}
				else if ( iswpunct( Format[i + 1] ) ) {
					ch = Format[i + 1];
					i++;
				}
			}

			if ( ch && ( j < ( MAX_ARG_LEN - 1 ) ) ) {
				temp[j++] = ch;
				temp[j] = '\0';
			}
		}

		if ( DisplayFlags & FF_EXPAND ) {
			ExpandVariables( temp, 1 );

			if ( DisplayFlags & FF_EXPAND_SAFELY )
				UnsafeChars( temp );
		}

		QPuts( temp );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;

		LineNumber++;
	}


	CloseMMFile( &MMFile );
	SetEnvironmentVariable( VarName, OldLineVar[0] ? OldLineVar : NULL );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	SetEnvironmentVariable( VarName, OldLineVar[0] ? OldLineVar : NULL );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI ffields( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t FormatString[MAX_ARG_LEN]	= L"";
	wchar_t LineStringSave[MAX_ARG_LEN]	= L"";

	unsigned int ParseFlags = PARSE_RETAIN_QUOTES;
	int DisplayFlags		= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	FileFieldsArgs Args		= { NULL, 0, NULL, 0, 0, 0 };
	LPTSTR LineString		= NULL;
	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_NO_TRIM | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( FFieldsHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( FFIELDS_COMMAND_VER ) );
			}
			else if ( !wcscmp( arg[i], L"/$" ) )
				DisplayFlags |= FF_DEBUG;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				ParseFlags |= PARSE_BREAK_COMMAS;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				DisplayFlags |= FF_DETAB;
			else if ( !_wcsicmp( arg[i], L"/E" ) ) {
				ParseFlags |= PARSE_EQUALS_KLUDGE;
				DisplayFlags |= FF_STRIP_EQUALS;
			}
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				wcscpy_s( FormatString, MAX_ARG_LEN, arg[i] + j );
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				DisplayFlags |= FF_HEADER;
			else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, (unsigned int*) &( Args.Columns ), NULL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad columns value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( Args.Columns && ( ( Args.Columns < MIN_COLUMNS ) || ( Args.Columns > MAX_COLUMNS ) ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Columns must be %u to %u: %u\r\n", MIN_COLUMNS, MAX_COLUMNS, Args.Columns );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( LineStringSave, MAX_ARG_LEN, arg[i] + j );
				LineString = LineStringSave;
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				ParseFlags &= ~PARSE_RETAIN_QUOTES;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				ParseFlags |= PARSE_BREAK_TABS;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				ParseFlags |= PARSE_FORCE_UPPERCASE;
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				ParseFlags |= PARSE_BREAK_SPACES;
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				DisplayFlags |= ( FF_EXPAND | FF_EXPAND_SAFELY );
			else if ( !_wcsicmp( arg[i], L"/XX" ) )
				DisplayFlags |= FF_EXPAND;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( _wcsicmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( ParseFlags & PARSE_BREAK_SPACES )
		ParseFlags &= ~PARSE_BREAK_TABS;

	if ( ( ParseFlags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE | PARSE_BREAK_TABS ) ) == 0 )
		ParseFlags |= PARSE_BREAK_SPACES;

	if ( ParseFlags & PARSE_BREAK_TABS )
		DisplayFlags &= ~FF_DETAB;


	if ( ( Args.Columns < MIN_COLUMNS ) || ( Args.Columns > MAX_COLUMNS ) ) {
		Args.Columns = EnvironmentNumericValue( L"COLUMNS", MIN_COLUMNS, MAX_COLUMNS );

		if ( ( Args.Columns < MIN_COLUMNS ) || ( Args.Columns > MAX_COLUMNS ) ) {
			if ( ConsoleHandle( STD_OUTPUT_HANDLE ) )
				Args.Columns = ConsoleWidth();
			else
				Args.Columns = DEFAULT_COLUMNS;

			if ( ( Args.Columns < MIN_COLUMNS ) || ( Args.Columns > MAX_COLUMNS ) )
				Args.Columns = DEFAULT_COLUMNS;
		}
	}

	if ( DisplayFlags & FF_DEBUG )
		Printf( L"   * Width = %d columns\n", Args.Columns );


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	//  Copy arguments list into separate variables, since the originals will be rewritten by FileFields() :

	wchar_t Filename[MAX_ARGS][MAX_ARG_LEN];
	int NumFilenames = 0;

	for ( int i = 0; i < numargs; i++ )
		if ( arglen[i] )
			if ( arg[i][0] != '/' )
				wcscpy_s( Filename[NumFilenames++], MAX_ARG_LEN, arg[i] );


	Args.Format			= FormatString;
	Args.ParseFlags		= ParseFlags;
	Args.LineString		= LineString;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = FileFields( (LPTSTR) StdinFilename, DisplayFlags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < NumFilenames; i++ ) {
		fhrv = HandleMatchingFiles( Filename[i], DisplayFlags, &Args, FileFields, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
		if ( fhrv == USER_ABORT )
			goto UserAbort;
		else if ( fhrv )
			rv = 2;
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	FFIELDS_COMMAND_VER


/*

	-			2021-10-15
				Added /T option to break at unquoted tabs.  Fixed a problem with FFAppendArg() trying
				to insert nonexistent args.

	-			2021-10-18
				Added /W option to set output line length.  Fixed a bug which could cause lines to be
				inadvertently detabbed.  Tweaked the routine which generates default format strings.

	-			2021-10-20
				Further tweaks to default format strings.

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				FileFields() filename is defined as a constant:  fix for mmfiles 1.4.4.2

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
